// system header
#include <algorithm>
#include <climits>
#include <cstring>
#include <fstream>
#include <iostream>
#include <memory>
#include <sstream>
#include <string>
#include <vector>
#include <sys/stat.h>
#include <unistd.h>

// library header
#ifdef QT_UI
    #include <QMessageBox>
#endif
#include <zsutil.h>

// AppImageKit includes
extern "C" {
    #include "appimage/appimage_shared.h"
}

// local header
#include "util/util.h"

namespace appimage::update::util {

    void removeNewlineCharacters(std::string& str) {
        str.erase(std::remove(str.begin(), str.end(), '\n'), str.end());
    }

    bool ltrim(std::string& s, char to_trim) {
        // TODO: find more efficient way to check whether elements have been removed
        size_t initialLength = s.length();
        s.erase(s.begin(), std::find_if(s.begin(), s.end(), [to_trim](int ch) {
            return ch != to_trim;
        }));
        return s.length() < initialLength;
    }

    bool rtrim(std::string& s, char to_trim) {
        // TODO: find more efficient way to check whether elements have been removed
        auto initialLength = s.length();
        s.erase(std::find_if(s.rbegin(), s.rend(), [to_trim](int ch) {
            return ch != to_trim;
        }).base(), s.end());
        return s.length() < initialLength;
    }

    bool trim(std::string& s, char to_trim) {
        // returns true if either modifies s
        auto ltrim_result = ltrim(s, to_trim);
        return rtrim(s, to_trim) && ltrim_result;
    }

    std::vector<std::string> split(const std::string& s, char delim) {
        std::vector<std::string> result;

        std::stringstream ss(s);
        std::string item;

        while (std::getline(ss, item, delim)) {
            result.push_back(item);
        }

        return result;
    }

    std::string join(const std::vector<std::string>& strings, const std::string& delim) {
        if (strings.empty()) {
            return {};
        }

        std::stringstream oss;
        oss << strings.front();

        for (auto it = strings.begin() + 1; it != strings.end(); ++it) {
            oss << delim << *it;
        }

        return oss.str();
    }

    std::string toLower(std::string s) {
        std::transform(s.begin(), s.end(), s.begin(), [](unsigned char c) { return std::tolower(c); });
        return s;
    }

    bool toLong(const std::string& str, long& retval, int base) {
        char* end = nullptr;
        const auto* cstr = str.c_str();

        auto rv = std::strtol(cstr, &end, base);
        if (errno == ERANGE || cstr == end || retval > LONG_MAX || retval < LONG_MIN)
            return false;

        retval = rv;
        return true;
    }

    bool isFile(const std::string& path) {
        std::ifstream ifs(path);
        return (bool) ifs && ifs.good();
    }

    void copyPermissions(const std::string& oldPath, const std::string& newPath) {
        mode_t oldPerms;
        auto errCode = zsync2::getPerms(oldPath, oldPerms);

        if (errCode != 0) {
            std::ostringstream ss;
            ss << "Error calling stat(): " << strerror(errCode);
#ifdef FLTK_UI
            fl_message("%s", ss.str().c_str());
#endif
#ifdef QT_UI
            QMessageBox::critical(nullptr, "Error", QString::fromStdString(ss.str()), QMessageBox::Close);
#endif
            exit(1);
        }

        chmod(newPath.c_str(), oldPerms);
    }

    void runApp(const std::string& path) {
        // make executable
        mode_t newPerms;
        auto errCode = zsync2::getPerms(path, newPerms);

        if (errCode != 0) {
            std::ostringstream ss;
            ss << "Error calling stat(): " << strerror(errCode);
#ifdef FLTK_UI
            fl_message("%s", ss.str().c_str());
#endif
#ifdef QT_UI
            QMessageBox::critical(nullptr, "Error", QString::fromStdString(ss.str()), QMessageBox::Close);
#endif
            exit(1);
        }

        chmod(path.c_str(), newPerms | S_IXUSR);

        // full path to AppImage, required for execl
        char* realPathToAppImage;
        if ((realPathToAppImage = realpath(path.c_str(), nullptr)) == nullptr) {
            auto error = errno;
            std::ostringstream ss;
            ss << "Error resolving full path of AppImage: code " << error << ": " << strerror(error) << std::endl;
#ifdef FLTK_UI
            fl_message("%s", ss.str().c_str());
#endif
#ifdef QT_UI
            QMessageBox::critical(nullptr, "Error", QString::fromStdString(ss.str()), QMessageBox::Close);
#endif
            exit(1);
        }

        if (fork() == 0) {
            putenv(strdup("STARTED_BY_APPIMAGEUPDATE=1"));

            std::cerr << "Running " << realPathToAppImage << std::endl;

            // make sure to deactivate updater contained in the AppImage when running from AppImageUpdate
            execl(realPathToAppImage, realPathToAppImage, nullptr);

            // execle should never return, so if this code is reached, there must be an error
            auto error = errno;
            std::cerr << "Error executing AppImage " << realPathToAppImage << ": code " << error << ": "
                      << strerror(error) << std::endl;
            exit(1);
        }
    }

    std::string readElfSection(const std::string& filePath, const std::string& sectionName) {
        unsigned long offset = 0, length = 0;

        auto rv = appimage_get_elf_section_offset_and_length(filePath.c_str(), sectionName.c_str(), &offset, &length);

        if (!rv || offset == 0 || length == 0)
            return "";

        std::ifstream ifs(filePath);
        ifs.seekg(offset);

        std::vector<char> buffer(length+1, 0);
        ifs.read(buffer.data(), length);

        return buffer.data();
    }

    std::string findInPATH(const std::string& name) {
        const std::string PATH = getenv("PATH");

        for (const auto& path : split(PATH, ':')) {
            std::ostringstream oss;
            oss << path << "/" << name;

            auto fullPath = oss.str();

            // TODO: check whether file is actually executable
            if (isFile(fullPath))
                return fullPath;
        }

        return "";
    }

    bool stringStartsWith(const std::string& string, const std::string& prefix) {
        return strncmp(string.c_str(), prefix.c_str(), prefix.size()) == 0;
    }

    std::string abspath(const std::string& path) {
        char* fullPath = nullptr;

        if ((fullPath = realpath(path.c_str(), nullptr)) == nullptr) {
            auto error = errno;
            std::cerr << "Failed to resolve full path to AppImage: " << strerror(error) << std::endl;
            return "";
        }

        std::string rv = fullPath;

        // clean up
        free(fullPath);
        fullPath = nullptr;

        return rv;
    }

    std::string pathToOldAppImage(const std::string& oldPath, const std::string& newPath) {
        if (oldPath == newPath) {
            return newPath + ".zs-old";
        }

        return abspath(oldPath);
    }

    std::string ailfsRealpath(const std::string& path) {
        std::stringstream ailfsBasePath;
        ailfsBasePath << "/run/user/" << getuid() << "/appimagelauncherfs/";

        if (path.find(ailfsBasePath.str()) == std::string::npos)
            return path;

        std::stringstream mapFilePath;
        mapFilePath << ailfsBasePath.str() << "/map";

        std::ifstream ifs(mapFilePath.str());

        if (!ifs)
            throw std::runtime_error("Could not open appimagelauncherfs map file");

        std::string pathFileName;
        {
            std::unique_ptr<char> pathCStr(strdup(path.c_str()));
            pathFileName = basename(pathCStr.get());
        }

        std::string currentLine;
        while (std::getline(ifs, currentLine)) {
            const std::string delim = " -> ";
            const auto delimiterPos = currentLine.find(delim);

            const auto ailfsFileName = currentLine.substr(0, delimiterPos);
            const auto targetFilePath = currentLine.substr(delimiterPos + delim.length());

            if (ailfsFileName == pathFileName)
                return targetFilePath;
        }

        throw std::runtime_error("Could not resolve path in appimagelauncherfs map file");
    }

    std::vector<char> makeBuffer(const std::string& str) {
        std::vector<char> buffer;
        buffer.resize(str.size());
        std::copy(str.begin(), str.end(), buffer.begin());
        buffer.emplace_back('\0');
        return buffer;
    }
}
